/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/**
 * @defgroup genedrop genedrop
 * @ingroup GDrop
 * @{
 */

#define PNAM "genedrop"
#include "genedrop.h"

int main (int argc, char *argv[])
{
   atexit (print_tail);                 /* initialize exit procedure          */
#if 0
   PCleanUp = clean_drop;               /* note:  code to free up memory      */
   atexit (clean_morg);                 /* at exit needs is not current       */
#endif

   print_head (*argv);

   if (!proc_pars (argc, argv, PNAM))   /* process input parameter statements */
      exit (EXIT_FAILURE);

   if (!proc_drop ())                   /* check parameter statements for     */
      exit (EXIT_FAILURE);              /* completeness & consistency         *
                                         * & prepare controls for program     */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);              /* set up seeds                       */

   if (!proc_fils ())                   /* set up file parameters             */
      exit (EXIT_FAILURE);

   if (!chk_pedfil ())                  /* read pedigree file, check errors   */
      give_up (ERR_qit);                /* & build the key table              *
                                         * sorts key table by component &     *
                                         * chronological order                */

   allocGenes ();                       /* allocates arrays for geno data     */

   allocPheno ();                       /* allocates arrays for pheno data    */

   setSeeds (MARK);                     /* sets seeds for markers             */

   if (LabelsOut)                       /* assigns gene labels to founders    */
      labelFounderGenes ();

   dropMarkerGenes ();                  /* assigns genes (and optionally      *
                                         * labels) to markers                 */

   getSeeds (MARK);                     /* saves marker seeds,                */
   assgnSeeds (TRAIT);                  /* sets trait seeds                   */
   setSeeds (TRAIT);

   dropTLocGenes ();                    /* assigns genes (and optionally      *
                                         * labels) to linked traits,          *
                                         * then to unlinked traits            */

   computTraits ();                     /* computes quantitative traits       */

   getSeeds (TRAIT);                    /* saves trait seeds                  */

   if (!ChronOut)                       /* if output order is same as         */
      sortIx ();                        /* input order, resorts               */

   openSimPed ();                       /* opens the output file              */

   writSimPed ();                       /* writes out the pedigree file       */
   fclose (oPedF);

   if (MarkSeedsOut || TraitSeedsOut)   /* if seeds are to be output,         */
   {
      openSeedOut ();                   /* writes to file                     */
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   freeStuff ();                        /* frees data structures              */

   exit (EXIT_SUCCESS);
}

/*. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

/**
 * allocates structures for genotypic data:  for each individual, 
 * for each locus, allocates k's, labels and alleles.
 */
void allocGenes (void)
{
   pKDat = malloc (nInd * NLoci * sizeof (KDAT));
   alloc_chk (WHERE, pKDat);

   if (LabelsOut)
   {
      pLablDat = malloc (nInd * NLoci * sizeof (LABL));
      alloc_chk (WHERE, pLablDat);
   }

   pGenoDat = malloc (nInd * NLoci * sizeof (GENO));
   alloc_chk (WHERE, pGenoDat);

   return;
}

/**
 * allocates structure for phenotypic data for each trait.
 */
void allocPheno (void)
{
   if (NTraits)
   {
      pPhenoDat = malloc (nInd * NTraits * sizeof (PHENO));
      alloc_chk (WHERE, pPhenoDat);
   }

   return;
}

/**
 * assigns paternal and maternal labels to each founder.
 */
void labelFounderGenes (void)
{
   KEYS_PP ppK;
   LABL_PTR pL;
   int   iLoc;
   int   iFound = 0;

   for (ppK = ppKeyTab; ppK < ppKeyTab + nInd; ++ppK) /* for ea individual    */
   {
      pL = pLablDat + ((*ppK)->ix - 1) * NLoci;

      if ((*ppK)->founder)              /* for each founder                   */
      {
         ++iFound;
         iLoc = 0;
         while (iLoc < NLoci)           /* for each locus                     */
         {
            pL->lablMa = iFound * 2;    /* assigns maternal label             */
            pL->lablPa = iFound * 2 - 1; /* assigns paternal label            */
            ++pL;
            ++iLoc;
         }
      }
   }
   return;
}

/**
 * 
 */
void dropMarkerGenes (void)
{
   KEYS_PP ppK;
   KDAT_PTR pKD;
   GENO_PTR pG;
   LABL_PTR pL;
   RMAP_PTR pR;
   ALLELE_PTR pA;
   int   iLoc;

   pL = NULL;

   if (NMarks)                          /* markers first ...                  */
   {
      for (ppK = ppKeyTab; ppK < ppKeyTab + nInd; ++ppK) /* for ea individual */
      {
         pKD = pKDat + ((*ppK)->ix - 1) * NLoci;
         pG = pGenoDat + ((*ppK)->ix - 1) * NLoci;
         if (LabelsOut)
            pL = pLablDat + ((*ppK)->ix - 1) * NLoci;
         pR = PRecombMap;
         pA = PAlleleFreq;
         iLoc = 0;

         while (iLoc < NLoci)           /* for each locus                     */
         {
            if (pR->recombTyp == FM || pR->recombTyp == OM)
            {
               if ((*ppK)->founder)     /* assigns marker genes to founders   */
                  foundGenes (pA, pKD, pG, iLoc);
               else
               {
                  inheritGenes (ppK, pR, pKD, pG, iLoc); /* and descendants   */
                  if (LabelsOut)
                     inheritLabels (ppK, pKD, pL, iLoc);
               }
            }
            ++pG;
            if (LabelsOut)
               ++pL;
            ++pR;
            ++pA;
            ++iLoc;
         }
      }
   }
   return;
}

void dropTLocGenes (void)
{
   KEYS_PP ppK;
   KDAT_PTR pKD;
   GENO_PTR pG;
   LABL_PTR pL;
   RMAP_PTR pR;
   ALLELE_PTR pA;
   int   iLoc;

   pL = NULL;

   if (NTLocs)                          /* Now tlocs ...                      */
   {
      for (ppK = ppKeyTab; ppK < ppKeyTab + nInd; ++ppK) /* for ea individual */
      {
         pKD = pKDat + ((*ppK)->ix - 1) * NLoci;
         pG = pGenoDat + ((*ppK)->ix - 1) * NLoci;
         if (LabelsOut)
            pL = pLablDat + ((*ppK)->ix - 1) * NLoci;
         pR = PRecombMap;
         pA = PAlleleFreq;
         iLoc = 0;

         while (iLoc < NLoci)           /* for each locus                     */
         {
            if (pR->recombTyp == FT || pR->recombTyp == IT
                || pR->recombTyp == ET || pR->recombTyp == UT)
            {
               if ((*ppK)->founder)     /* assigns trait genes                */
                  foundGenes (pA, pKD, pG, iLoc); /* to founders              */
               else
               {
                  inheritGenes (ppK, pR, pKD, pG, iLoc); /* and descendents   */
                  if (LabelsOut)
                     inheritLabels (ppK, pKD, pL, iLoc);
               }
            }
            ++pG;
            if (LabelsOut)
               ++pL;
            ++pR;
            ++pA;
            ++iLoc;
         }
      }
   }
   return;
}

/**
 * computes additive and residual variances and quantitative trait value
 * for each member at each trait locus.
 * (traits are computed in numerical order)
 */
void computTraits (void)
{
   KEYS_PP ppK;
   GENO_PTR pG;
   PHENO_PTR pPh;
   TPARS_PTR pTr;
   int   iTrait;

   if (!NTraits)
      return;

   for (ppK = ppKeyTab; ppK < ppKeyTab + nInd; ++ppK) /* for ea individual    */
   {
      pG = pGenoDat + ((*ppK)->ix - 1) * NLoci;
      pPh = pPhenoDat + ((*ppK)->ix - 1) * NTraits;
      pTr = PTraitPars;

      for (iTrait = 0; iTrait < NTraits; ++iTrait) /* for each trait          */
      {
         if ((*ppK)->founder)
            foundTrait (pPh, pTr);
         else
            inheritTrait (ppK, pPh, pTr, iTrait);

         makQuant (pG, pPh, pTr);

         ++pPh;
         ++pTr;
      }
   }
   return;
}

void openSimPed (void)
{
   if (!(oPedF = open_fil (OPED)))
      give_up (ERR_qit);

   return;
}

void writSimPed (void)
{
   KEYS_PP ppK = ppKeyTab;
   PED_PTR pP;
   PEDI_PTR pPI;
   GENO_PTR pG;
   LABL_PTR pL;
   PHENO_PTR pPh;
   RMAP_PTR pR = PRecombMap;
   TPARS_PTR pT = PTraitPars;
   TLOUT_PTR pTL = PTLocOut;

   int   i, j, ixInd, widLab;

   for (j = 0; j < NNInts; ++j)         /* gets no. characters for writing    */
      widPInt[j] = intWid (minPInt[j], maxPInt[j]);
   widLab = intWid (1, nFound * 2);

   /*for each individual:  */
   for (i = 0; i < nInd; ++i)
   {
      ixInd = (*ppK)->ix - 1;           /* gets index for next individual     */

      pP = pPedDat + ixInd;             /* outputs standard integers (name,
                                         * parents, gender, and if required
                                         * observed indicator)                */
      for (j = 0; j < NStdOut; ++j)
      {
         switch (StdOut[j])
         {
             case NAM:
                {
                   fprintf (oPedF, "%*s", widPNam, pP->nam);
                   break;
                }
             case PA:
                {
                   fprintf (oPedF, " %*s", widPNam, pP->namPa);
                   break;
                }
             case MA:
                {
                   fprintf (oPedF, " %*s", widPNam, pP->namMa);
                   break;
                }
             case SEX:
                {
                   fprintf (oPedF, " %2d", pP->sex);
                   break;
                }
             case OBS:
                {
                   fprintf (oPedF, " %d", pP->obsrvd);
                   break;
                }
             default:
                break;
         }
      }
      pPI = pIntDat + NNInts * ixInd;   /* outputs integers, if input         */

      for (j = 0; j < NNInts; ++j)
         fprintf (oPedF, " %*d", widPInt[j], *(pPI + j));

      fprintf (oPedF, " ");

      if (LabelsOut)                    /* optionally outputs marker labels   */
      {
         pL = pLablDat + NLoci * ixInd;
         pR = PRecombMap;

         for (j = 0; j < NLoci; ++j)
         {
            if (pR->recombTyp == FM || pR->recombTyp == OM)
               if (!ObsOut && !pP->obsrvd)
                  fprintf (oPedF, "  %*d %*d", widLab, MISSLABEL,
                           widLab, MISSLABEL);
               else
                  fprintf (oPedF, "  %*d %*d",
                           widLab, pL->lablPa, widLab, pL->lablMa);
            ++pL;
            ++pR;
         }
      }

      pG = pGenoDat + NLoci * ixInd;    /* outputs marker genotypes           */
      pR = PRecombMap;

      for (j = 0; j < NLoci; ++j)
      {
         if (pR->recombTyp == FM || pR->recombTyp == OM)
            if (!ObsOut && !pP->obsrvd)
               fprintf (oPedF, "  %*d %*d", WidAlle, MISSGENE,
                        WidAlle, MISSGENE);
            else if (pG->genoPa < pG->genoMa)
               fprintf (oPedF, "  %*d %*d",
                        WidAlle, pG->genoPa + 1, WidAlle, pG->genoMa + 1);
            else
               fprintf (oPedF, "  %*d %*d",
                        WidAlle, pG->genoMa + 1, WidAlle, pG->genoPa + 1);
         ++pG;
         ++pR;
      }

      if (NTLocs && (LabelsOut || LatentOut))
         fprintf (oPedF, " ");

      if (NTLocs > 0 && LabelsOut)      /* optionally outputs tloc labels     */
      {
         pTL = PTLocOut;
         for (j = 0; j < NTLocs; ++j)
         {
            pL = pLablDat + NLoci * ixInd + pTL->ixG;
            if (!ObsOut && !pP->obsrvd)
               fprintf (oPedF, "  %*d %*d", widLab, MISSLABEL,
                        widLab, MISSLABEL);
            else
               fprintf (oPedF, "  %*d %*d",
                        widLab, pL->lablPa, widLab, pL->lablMa);
            ++pTL;
         }
      }

      if (NTLocs && LatentOut)          /* optionally outputs tloc genotypes  */
      {
         pTL = PTLocOut;
         for (j = 0; j < NTLocs; ++j)
         {
            pG = pGenoDat + NLoci * ixInd + pTL->ixG;
            if (!ObsOut && !pP->obsrvd)
               fprintf (oPedF, "  %*d %*d", WidAlle, MISSGENE,
                        WidAlle, MISSGENE);

            else if (pG->genoPa < pG->genoMa)
               fprintf (oPedF, "  %*d %*d", WidAlle, pG->genoPa + 1,
                        WidAlle, pG->genoMa + 1);
            else
               fprintf (oPedF, "  %*d %*d", WidAlle, pG->genoMa + 1,
                        WidAlle, pG->genoPa + 1);
            ++pTL;
         }
      }

      if (NTraits)
         fprintf (oPedF, " ");

      pT = PTraitPars;

      for (j = 0; j < NTraits; ++j)     /* outputs quantitative data          */
      {
         pPh = pPhenoDat + NTraits * ixInd + pT->ixPh;

         if (LatentOut)                 /* optionally outputs variances       */
         {
            if (!ObsOut && !pP->obsrvd)
               fprintf (oPedF, " %*.*f %*.*f %*d",
                        pT->wid, pT->dec, 0.,
                        pT->wid, pT->dec, 0., pT->wid, MISSING);
            else
               fprintf (oPedF, " %*.*f %*.*f %*.*f",
                        pT->wid, pT->dec, pPh->addi,
                        pT->wid, pT->dec, pPh->resi,
                        pT->wid, pT->dec, pPh->quant);
         }
         else if (!ObsOut && !pP->obsrvd)
            fprintf (oPedF, " %*d", pT->wid, MISSING);
         else
            fprintf (oPedF, " %*.*f", pT->wid, pT->dec, pPh->quant);
         ++pT;
      }

      fprintf (oPedF, "\n");            /* ends current record              */
      fflush (oPedF);

      ++ppK;
   }
   return;
}

void openSeedOut (void)
{
   if (!(oSeedF = open_fil (OSEED)))
      give_up (ERR_qit);

   return;
}

void freeStuff (void)
{
   freeKey ();

   free (pPedDat);
   pPedDat = NULL;

   if (NNInts)
   {
      free (pIntDat);
      pIntDat = NULL;
   }

   free (pGenoDat);
   pGenoDat = NULL;

   if (pLablDat)
   {
      free (pLablDat);
      pLablDat = NULL;
   }

   if (NTraits)
   {
      free (pPhenoDat);
      pPhenoDat = NULL;
   }

   free (pKDat);
   pKDat = NULL;

   free_fnams ();

   return;
}

/**
 * assigns a pair of alleles to the founder at the specified locus.
 */
void foundGenes (ALLELE_PTR pA, KDAT_PTR pKD, GENO_PTR pG, int iLoc)
{
   int   j;
   double rM, rP;

   rM = rans ();
   rP = rans ();

   for (j = 0; j < pA->nAlleles && rM > *((pA->pCumFreq) + j); ++j)
      ;
   pG->genoMa = j;                      /* selects maternal allele          */

   for (j = 0; j < pA->nAlleles && rP > *((pA->pCumFreq) + j); ++j)
      ;
   pG->genoPa = j;                      /* selects paternal allele          */

   (pKD + iLoc)->kMat = FOUNDER;
   (pKD + iLoc)->kPat = FOUNDER;

   return;
}

/**
 * for the non-founder, determines the maternal and paternal inheritance
 * at the specified locus.
 * saves the pair of k's (inheritance indicators) and the pair of alleles.
 */
void inheritGenes (KEYS_PP ppK, RMAP_PTR pR, KDAT_PTR pKD, GENO_PTR pG,
                   int iLoc)
{
   int   kM, kP, kA, kB;                /* assigns k's and thus alleles     */
   double rM, rP;                       /* to nonfounders                   */

   rM = rans ();
   rP = rans ();

   if (pR->recombTyp == FM || pR->recombTyp == UT)
   {
      if (rM < .5)                      /* first marker on chromosome       */
         kM = GM;                       /* or unlinked trait                */
      else
         kM = GF;

      if (rP < .5)
         kP = GM;
      else
         kP = GF;
   }

   else if (pR->recombTyp == OM || pR->recombTyp == ET)
   {
      kM = (pKD + pR->iA)->kMat;        /* is not first marker on chrom     */
      if (rM < pR->probA[FEM])          /* or, is trait with preceding      */
         kM = !kM;                      /* locus & no following marker      */

      kP = (pKD + pR->iA)->kPat;
      if (rP < pR->probA[MAL])
         kP = !kP;
   }

   else if (pR->recombTyp == FT)
      /* trait is first locus on chrom */
   {
      kM = (pKD + pR->iB)->kMat;
      if (rM < pR->probB[FEM])
         kM = !kM;

      kP = (pKD + pR->iB)->kPat;
      if (rP < pR->probB[MAL])
         kP = !kP;
   }

   else if (pR->recombTyp == IT)
   {
      kA = (pKD + pR->iA)->kMat;        /* trait has preceding locus        */
      kB = (pKD + pR->iB)->kMat;        /* and following marker             */
      if (kA == kB)
      {
         if (rM < pR->probA[FEM])
            kM = !kA;
         else
            kM = kA;
      }
      else
      {
         if (rM < pR->probB[FEM])
            kM = kA;
         else
            kM = kB;
      }

      kA = (pKD + pR->iA)->kPat;
      kB = (pKD + pR->iB)->kPat;
      if (kA == kB)
      {
         if (rP < pR->probA[MAL])
            kP = !kA;
         else
            kP = kA;
      }
      else
      {
         if (rP < pR->probB[MAL])
            kP = kA;
         else
            kP = kB;
      }
   }

   (pKD + iLoc)->kMat = kM;

   if (kM == GM)
      pG->genoMa = (pGenoDat + ((*ppK)->ixMa - 1) * NLoci + iLoc)->genoMa;
   else if (kM == GF)
      pG->genoMa = (pGenoDat + ((*ppK)->ixMa - 1) * NLoci + iLoc)->genoPa;
   else
      give_up_where (WHERE, ERR_inh, "inheritGenes", (*ppK)->nam);

   (pKD + iLoc)->kPat = kP;

   if (kP == GM)
      pG->genoPa = (pGenoDat + ((*ppK)->ixPa - 1) * NLoci + iLoc)->genoMa;
   else if (kP == GF)
      pG->genoPa = (pGenoDat + ((*ppK)->ixPa - 1) * NLoci + iLoc)->genoPa;
   else
      give_up_where (WHERE, ERR_inh, "inheritGenes", (*ppK)->nam);

   return;
}

/**
 * for the non-founder, assigns a pair of labels, based on the 
 * at the specified locus.
 * saves the pair of k's (inheritance indicators) and the pair of alleles.
 */
void inheritLabels (KEYS_PP ppK, KDAT_PTR pKD, LABL_PTR pL, int iLoc)
{
   int   kM, kP;

   kM = (pKD + iLoc)->kMat;
   kP = (pKD + iLoc)->kPat;

   if (kM == GM)                        /* inherit label from mat GM or GP    */
      pL->lablMa = (pLablDat + ((*ppK)->ixMa - 1) * NLoci + iLoc)->lablMa;
   else if (kM == GF)
      pL->lablMa = (pLablDat + ((*ppK)->ixMa - 1) * NLoci + iLoc)->lablPa;
   else
      give_up_where (WHERE, ERR_inh, "inheritLabels", (*ppK)->nam);

   if (kP == GM)                        /* inherit label from pat Gm or GP    */
      pL->lablPa = (pLablDat + ((*ppK)->ixPa - 1) * NLoci + iLoc)->lablMa;
   else if (kP == GF)
      pL->lablPa = (pLablDat + ((*ppK)->ixPa - 1) * NLoci + iLoc)->lablPa;
   else
      give_up_where (WHERE, ERR_inh, "inheritLabels", (*ppK)->nam);

   return;
}

void foundTrait (PHENO_PTR pPh, TPARS_PTR pTP)
{
   pPh->addi = normal (0., pTP->addi_term_found);

   return;
}
void inheritTrait (KEYS_PP ppK, PHENO_PTR pPh, TPARS_PTR pTP, int iT)
{
   double addiMa, addiPa;

   addiMa = (pPhenoDat + ((*ppK)->ixMa - 1) * NTraits + iT)->addi;
   addiPa = (pPhenoDat + ((*ppK)->ixPa - 1) * NTraits + iT)->addi;
   pPh->addi = normal ((addiMa + addiPa) / 2., pTP->addi_term);

   return;
}

void makQuant (GENO_PTR pG, PHENO_PTR pPh, TPARS_PTR pTP)
{
   int   i;

   pPh->resi = normal (0., pTP->resi_term); /*               0 for 0,0        */
   i = (pG + pTP->ixG)->genoMa + (pG + pTP->ixG)->genoPa; /* 1 for 0,1 or 1,0 */
   pPh->quant = pTP->means[i] + pPh->addi + pPh->resi; /*    2 for 1,1        */

   return;
}

/** @} */
