/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/**
 * this function builds global data structures used by the Autozyg programs.
 * 
 * proc_auto() has as its input the results of parsing the parameter statements
 * for the Genedrop programs ibddrop and markerdrop, and for the Autozyg
 * programs, lm_auto, lm_ibdtests, lm_map and lm_pval.
 * proc_auto() uses many of the proc_drop() data structures and functions.
 * (Think of proc_auto as an extension of proc_drop.)
 * This function checks parameter statement data for consistency and complete-
 * ness and, if there are no errors, builds the global data structures.
 *
 * One chromosome only is selected
 * At least 1 marker must be simulated (for ibddrop and markerdrop)
 * or selected (for lm_map, lm_ibdtests, lm_pval and lm_auto).
 * ibddrop simulates 1 tloc.
 *
 * Any combination of sex-not-specified maps or sex-specific maps may be used
 *
 *      MAJ  revs 5/2000, 1/2002, 8/2003, 6/2005, 6/2006, 5/2007
 */

int proc_auto (void)
{
   int   entry, sex, m, t_no, tl_no, loc_no;
   int   sxsp_maps = FALSE;             /* whether or not sex-spec maps used  */
   int   tloc_maps = FALSE;             /* whether or not there are tloc maps */
   SIM_PTR p_s;
   PRNTMAP_PTR p_m;
   PRDIST_PTR p_l;

   if (!test_auto_sel ())               /* checks simulate/select statement   */
      if (proc_err ())                  /* (fnc uses program names)           */
         return FALSE;

   if (n_err)
      return proc_end ();

   if (TRAIT_SELCTD && TRAIT_TLOC)      /* finds the tloc for the selected    */
      if (!find_trait_tloc ())          /* trait                              */
         if (proc_err ())
            return FALSE;

   if (INHER_SELCTD && INHER_TLOC)      /* finds the tloc for the selected x  */
      if (!find_inher_tloc ())          /* inheritance                        */
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (TLOC_MAP)                        /* set tloc maps flag     */
      if ((TRAIT_SELCTD && TRAIT_TLOC) || TLOC_SELCTD
          || (INHER_SELCTD && INHER_TLOC))
         tloc_maps = TRUE;

   if (TRAIT_SELCTD && tloc_maps)       /* finds the chrom map (including     */
      if (!find_tr_tloc_map ())         /* unlinked specs) for tloc for the   */
         if (proc_err ())               /* requested trait                    */
            return FALSE;

   if (TLOC_SELCTD && tloc_maps)        /* finds the chrom map (including     */
      if (!find_tloc_map ())            /* unlinked specs) for requested tloc */
         if (proc_err ())
            return FALSE;

   if (INHER_SELCTD && tloc_maps)       /* finds the chrom map (including     */
      if (!find_inh_tloc_map ())        /* unlinked specs) for tloc for the   */
         if (proc_err ())               /* requested inheritance              */
            return FALSE;

   if (n_err)
      return proc_end ();

   if (tloc_maps)                       /* if tloc,                           */
   {
      if (!updt_sim_lnkd_tloc ())       /* adds the mapped tloc to chromosome */
         if (proc_err ())               /* to be sim/sel                      */
            return FALSE;
      if (!updt_sim_unln_tloc ())       /* adds the unlinked tloc             */
         if (proc_err ())               /* to be sim/sel                      */
            return FALSE;
   }

   entry = FALSE;
   p_s = nxt_sim_chrom (&entry);        /* one chromosome only is selected    */

   if (MARK_MAP)                        /* checks sex-avrgd & sex-spcfc maps  */
   {
      if (!test_mark_map (p_s, &sxsp_maps)) /* checks marker map              */
         if (proc_err ())
            return FALSE;
      if (tloc_maps)                    /* checks tloc map                    */
         if (!test_tloc_map (p_s, &sxsp_maps))
            if (proc_err ())
               return FALSE;
   }

   if (tloc_maps)                       /* for the sim chromosome,            */
   {                                    /* if mapped tloc required,           */
      entry = FALSE;                    /* adds to tloc list                  */
      while (tl_no = nxt_sim_mapd_tl_no (&entry, p_s))
         mak_tloc_list (tl_no, LNKD_TLOC);
      entry = FALSE;                    /* if unlinked tloc required          */
      while (tl_no = nxt_sim_unln_tl_no (&entry)) /* adds to sim unlnkd list  */
         mak_tloc_list (tl_no, LNKD_TLOC);
   }

   if (MARK_ALLELE)                     /* for ea marker, check allele freqs  */
   {
      entry = FALSE;
      while (m = nxt_sim_mark_no (&entry, p_s))
         if (!test_mark_freq (p_s, m))
            if (proc_err ())
               return FALSE;
   }

   if (sim_traits && TLOC_ALLELE && !INHER_SELCTD) /* for tloc, check freqs   */
   {
      entry = FALSE;
      while (tl_no = nxt_sim_tl_no (&entry))
         if (!test_tloc_freq (tl_no))
            if (proc_err ())
               return FALSE;
   }

   entry = FALSE;
   t_no = nxt_sim_tr_no (&entry);

   if (sim_traits)                      /* establishes trait type             */
      if (!test_trait_typ ())
         if (proc_err ())
            return FALSE;

   if (sim_traits && TRT_TLC_GENO)      /* for ea trait/tloc                  */
   {                                    /* checks genotype means              */
      entry = FALSE;
      while (tl_no = nxt_tr_tl_no (&entry, t_no))
         if (!test_tr_tl_genotyp (t_no, tl_no))
            if (proc_err ())
               return FALSE;
   }

   if (sim_traits && TRAIT_VARI)        /* for ea trait, checks variances     */
      if (!test_trait_vars (t_no))
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (MARK_MAP)                        /* makes the "print map" for the      */
   {                                    /* selected markers and the tloc.    
                                         * assigns loci to sex-specific maps.
                                         * for maps not present,     
                                         * defaults to AVG, then F or M maps  */

      prnt_chrom_head (p_s);

      for (sex = FEM; sex <= AVG; ++sex) /* for ea gender,                    */
      {
         p_m = mak_markprmap (p_s, sex); /* for the selected markers          */

         if (MARK_MIN_DIST)
            if (!chk_markmindist (p_m)) /* checks dists between markers       */
               if (proc_err ())
                  return FALSE;

         if (tloc_maps)                 /* then, for sel tloc, if present     */
            mak_tlocprmap (p_s, p_m, sex);

         shrink_prntmap (p_m);          /* removes non-selected markers       */
         fix_prmap_dist (p_m);          /* finds dists betw adj loci for prnt */

         if (tloc_maps)                 /* for ea gender to be used           */
            if ((sxsp_maps && (sex == FEM || sex == MAL))
                || (!sxsp_maps && sex == AVG))
               if (!chk_tlocprmap (p_m)) /* checks tloc distances             */
                  if (proc_err ())
                     return FALSE;
      }
      prnt_map (sxsp_maps);             /* prints the maps                    */

      if (MARK_ALLELE)                  /* prints marker allele freqs         */
         prnt_mark_freqs (p_s);

      if (sim_traits && TLOC_ALLELE && !INHER_SELCTD) /* prints tloc freqs    */
         prnt_tloc_freqs (p_m);
   }

   if (tloc_maps)
      if (UNLTLOC_SELCTD)
      {
         prnt_unltloc_head ();          /* for the selected unmapped tloc,    */

         if (sim_traits && TLOC_ALLELE && !INHER_SELCTD)
            prnt_unltloc_freqs ();      /* prints freqs of unmapped loci      */
         else
            prnt_unltloc_list ();       /* or names of unmapped loci          */
      }

   if (INHER_SELCTD)                    /* print inher/tloc list              */
   {
      prnt_inher_head ();
      prnt_inh_tl_list ();
   }

   if (TRAIT_SELCTD)                    /* lists trait and prints params      */
   {
      prnt_trait_head ();

      if (TRT_TLC_GENO)
         prnt_tr_tl_genotyp ();
      else if (TRAIT_TLOC)
         prnt_tr_tl_list ();
      else if (!TRAIT_VARI)
         prnt_tr_list ();

      if (TRAIT_VARI)
         prnt_trait_vari ();
   }
   fprintf (stdout, "\n");

/*  FIXME ************ make specific for traits, tlocs and inher ** */
   NMarks = p_m->no_mark_sel;           /* counts loci                        */
   NLnkdTL = p_m->no_tloc_sel;
   if (UNLTLOC_SELCTD)
      NUnlnTL = sim_unlnkd_tlocs->no_tloc_sel;
   else
      NUnlnTL = 0;
   NTLocs = NLnkdTL + NUnlnTL;
   if (!sim_traits)
      NTraits = 0;
   else
      NTraits = sim_traits->no_traits;

   NLoci = NMarks + 1;                  /* now we count the tloc, 
                                         * whether or not one is used         */

   if (tloc_maps)
   {
      entry = FALSE;
      while (tl_no = nxt_sim_tl_no (&entry))
         mak_tlocnam (tl_no);           /* add tloc name                      */
   }

   alloc_marek ();                      /* stores marker recomb fracts in
                                         * Marek[][]                          */
   mak_marek ();

   mak_tarek ();                        /* stores tloc recomb fract and pos   *
                                         * (or .5, .5 if no linked tloc)      */

   entry = FALSE;                       /* for each locus, stores allele      */
   p_m = nxt_prntmap (&entry, FEM);     /* frequencies in Qq[][]              */

   if (MARK_ALLELE || TLOC_ALLELE)      /* first markers and linked tloc      */
   {
      entry = FALSE;
      while (p_l = nxt_prntmap_locus (&entry, p_m))
         if (p_l->locus_typ == TLOC_LOC && INHER_SELCTD)
            mak_allefrq (NULL, NULL, 0);
         else
            mak_allefrq (p_m, p_l, p_l->no);

      if (UNLTLOC_SELCTD)               /* or unlinked tloc                   */
      {
         entry = FALSE;
         while (tl_no = nxt_sim_unln_tl_no (&entry))
            mak_allefrq (NULL, NULL, tl_no);
      }
      else if (!NTLocs)                 /* or no tloc                         */
         mak_allefrq (NULL, NULL, 0);
   }

   if (MARK_NAME)
   {
      if (!test_marknam (p_s))          /* checks marker names                */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         if (mak_marknam (p_s))         /* stores and prints marker names     */
            prnt_marknam (p_s);         /* if provided                        */
      }
      else
         return proc_end ();
   }

   if (MARK_DATA)
   {
      if (!test_markdat (p_s))          /* checks marker data                 */
         if (proc_err ())
            return FALSE;
      if (!n_err)
      {
         mak_markdat (p_s);             /* stores and prints marker data      */
         prnt_markdat (p_s);
      }
      else
         return proc_end ();
   }

   if (PROBAND_GAM)
   {                                    /* checks proband gametes             */
      if (!test_probgam (p_s))          /* and locus window                   */
         if (proc_err ())
            return FALSE;

      if (PROB_GAM_WNDW)                /* checks locus window for scoring    */
         if (!test_probgam_jnt (p_s))   /* pair of gametes                    */
            if (proc_err ())
               return FALSE;

      if (PROB_GAM_PTRN)                /* checks locus window for scoring    */
         if (!test_probgam_pat (p_s))   /* pattern labels                     */
            if (proc_err ())
               return FALSE;
      if (!n_err)
      {
         mak_probgam ();                /* stores proband gametes             */
         mak_patlab ();                 /* stores window pattern labels       */
         prnt_probgam ();               /* prints locus window, proband       */
      }                                 /* gametes and window patterns        */
   }

   if (!test_auto_pars ())              /* checks MC computational parametrs */
      if (proc_err ())
         return FALSE;

   if (TRAIT_DATA)                      /* checks input record parmeters     */
      if (!test_ped_trtdat ())
         if (proc_err ())
            return FALSE;
   if (n_err)
      return proc_end ();

   mak_ped_dat ();

   if (TRAIT_DATA)                      /* sets up and reports trait ped data */
   {
      prnt_trt_typ ();

      prnt_ped_dat ();
   }

   if (TRAIT_PENE)                      /* checks incomplete penetrances      */
   {                                    /* for discete trait data             */
      if (!test_penetrnc ())
         if (proc_err ())
            return FALSE;

      prnt_penetrnc ();                 /* and prints penetrances             */
   }

   if (n_err)
      return proc_end ();

   if (NTraits && (TraitTyp == QUANTITATIVE)) /* stores geno means & variance */
   {
      alloc_traits ();
      entry = FALSE;
      tl_no = nxt_tr_tl_no (&entry, t_no);
      loc_no = get_tl_sim_loc_no (tl_no);
      mak_trait (t_no, tl_no, loc_no, 0);
/*      prnt_qtrait_pars (p_s);           prints means and variances         */
/*                    FIXME:  this routine does not exist: EAT 3/26/2008     */
   }

   prnt_auto_pars ();                   /* prints MC parameters               */

   return TRUE;
}
