/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/**
 * this function builds global data structures used by the Autozyg programs.
 * 
 * proc_lods() has as its input the results of parsing the parameter statements
 * for the for the LR_Lods programs, lm_lods and lm_schnell and for the
 * Lodscore programs, lm_bayes, lm_makers. proc_lods() uses many of the
 * same data structures and functions  as used by proc_drop() and proc_auto().
 * This function checks parameter statement data for consistency and complete-
 * ness and, if there are no errors, builds the global data structures for the
 * Morgan main programs and subroutine libraries.
 *
 * One chromosome only is selected
 * A set of test positions is assigned to the trait locus.
 *
 * Any combination of sex-not-specified maps or sex-specific maps may be used
 *
 *      MAJ  revs 5/2000, 1/2002, 8/2003, 6/2005, 6/2006, 10/2007
 */

int proc_lods (void)
{
   int   entry, sex, m, t_no, tl_no, loc_no;
   int   sxsp_maps = FALSE;             /* whether or not sex-spec maps used  */
   SIM_PTR p_s;
   PRNTMAP_PTR p_m;
   PRNTVMAP_PTR p_vm;
   PRDIST_PTR p_l;

   if (!test_lods_sel ())               /* IMPORGTANT: proc_lods expects      */
      if (proc_err ())                  /* 1 trait with varying test pos      */
         return FALSE;                  /* for a single trait locus           */

   if (n_err)
      return proc_end ();

   if (!find_trait_tloc ())             /* (TRAIT_SELCTD && TRAIT_TLOC)       */
      if (proc_err ())                  /* finds tloc for the selected trait  */
         return FALSE;                  /* (assumes 1 tloc)                   */

   if (n_err)
      return proc_end ();

   if (!find_tr_tloc_vmap ())           /* (TLOC_VMAP)                        */
      if (proc_err ())                  /* find the chrom map                 */
         return FALSE;                  /* for the test pos for the tloc      */

   if (n_err)
      return proc_end ();

   if (!updt_sim_vtloc ())              /* (TLOC_VMAP)                        */
      if (proc_err ())                  /* adds the tloc to chrom selectd     */
         return FALSE;

   entry = FALSE;
   p_s = nxt_sim_chrom (&entry);        /* one chromosome only is selected    */

   if (!test_mark_map (p_s, &sxsp_maps)) /* (MARK_MAP)                        */
      if (proc_err ())                  /* checks marker map                  */
         return FALSE;
   if (!test_tloc_vmap (p_s, &sxsp_maps)) /* (TLOC_VMAP)                      */
      if (proc_err ())                  /* checks tloc test positions         */
         return FALSE;
   if (!test_tloc_vext (p_s))
      if (proc_err ())
         return FALSE;
   if (!test_tloc_vint (p_s))
      if (proc_err ())
         return FALSE;

   entry = FALSE;                       /* (TLOC_VMAP)                        */
   while (tl_no = nxt_sim_mapd_tl_no (&entry, p_s)) /* adds the required      */
      mak_tloc_list (tl_no, LNKD_TLOC); /* tloc to the list                   */

   if (MARK_ALLELE)                     /* for ea marker, check allele freqs  */
   {
      entry = FALSE;
      while (m = nxt_sim_mark_no (&entry, p_s))
         if (!test_mark_freq (p_s, m))
            if (proc_err ())
               return FALSE;
   }

   if (TLOC_ALLELE)                     /* for the tloc, check allele freqs   */
   {
      entry = FALSE;
      while (tl_no = nxt_sim_tl_no (&entry))
         if (!test_tloc_freq (tl_no))
            if (proc_err ())
               return FALSE;
   }

   entry = FALSE;
   t_no = nxt_sim_tr_no (&entry);

   if (!test_trait_typ ())              /* (TRAIT_SELCTD)                     */
      if (proc_err ())                  /* checks trait type                  */
         return FALSE;

   if (TRT_TLC_GENO)                    /* checks genotype means              */
   {
      while (tl_no = nxt_tr_tl_no (&entry, t_no))
         if (!test_tr_tl_genotyp (t_no, tl_no))
            if (proc_err ())
               return FALSE;
   }

   if (TRAIT_VARI)                      /* checks additive, residual vari's   */
      if (!test_trait_vars (t_no))
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (MARK_MAP)                        /* makes the "print map" for the sel  */
   {                                    /* markers and tloc test positions. 
                                         * assigns loci to sex-specific maps.
                                         * for maps not present,     
                                         * defaults to AVG, then F or M maps  */
      prnt_chrom_head (p_s);

      for (sex = FEM; sex <= AVG; ++sex) /* for ea gender,                    */
      {
         p_m = mak_markprmap (p_s, sex); /* first, for the selected markers   */

         if (MARK_MIN_DIST)             /* checks dists betwn markers, if req */
            if (!chk_markmindist (p_m))
               if (proc_err ())
                  return FALSE;

         if (!mak_tlocvprmap (p_s, p_m, sex)) /* chcks test pos of slctd tloc */
            if (proc_err ())            /* (TLOC_VMAP)                        */
               return FALSE;

         shrink_prntmap (p_m);          /* removes non-sel marker loci and    */
         fix_prmap_dist (p_m);          /* finds dists betw adj mrks for prnt */
      }
      if (n_err)
         return proc_end ();

      prnt_map (sxsp_maps);             /* prints the maps                    */

      prnt_vmap (p_m->no_mark_sel, sxsp_maps); /* TLOC_VMAP                   */

      if (MARK_ALLELE)                  /* prints marker allele freqs         */
         prnt_mark_freqs (p_s);

      entry = FALSE;
      p_vm = nxt_tl_vmap (AVG, p_s);
      if (TLOC_ALLELE)                  /* prints tloc allele freqs           */
         prnt_vtloc_freqs (p_vm);
   }
   fprintf (stdout, "\n");

   if (TRAIT_SELCTD)                    /* lists traits and prints params     */
   {
      prnt_trait_head ();

      if (TRT_TLC_GENO)
         prnt_tr_tl_genotyp ();
      else if (TRAIT_TLOC)
         prnt_tr_tl_list ();
      else if (!TRAIT_VARI)
         prnt_tr_list ();

      if (TRAIT_VARI)
         prnt_trait_vari ();
   }
   fprintf (stdout, "\n");

#if 0
/*  FIXME ************ make specific for traits, tlocs and inher ** */
   NMarks = p_m->no_mark_sel;           /* counts loci                        */
   NLnkdTL = p_m->no_tloc_sel;
   if (UNLTLOC_SELCTD)
      NUnlnTL = sim_unlnkd_tlocs->no_tloc_sel;
   else
      NUnlnTL = 0;
   NTLocs = NLnkdTL + NUnlnTL;
   if (!sim_traits)
      NTraits = 0;
   else
      NTraits = sim_traits->no_traits;

   NLoci = NMarks + 1;                  /* now we count the tloc, 
                                         * whether or not one is used         */

   if (tloc_maps)
   {
      entry = FALSE;
      while (tl_no = nxt_sim_tl_no (&entry))
         mak_tlocnam (tl_no);           /* add tloc name                      */
   }

   alloc_marek ();                      /* stores marker recomb fracts in
                                         * Marek[][]                          */
   mak_marek ();

   mak_tarek ();                        /* stores tloc recomb fract and pos   *
                                         * (or .5, .5 if no linked tloc)      */

   entry = FALSE;                       /* for each locus, stores allele      */
   p_m = nxt_prntmap (&entry, FEM);     /* frequencies in Qq[][]              */

   if (MARK_ALLELE || TLOC_ALLELE)      /* first markers and linked tloc      */
   {
      entry = FALSE;
      while (p_l = nxt_prntmap_locus (&entry, p_m))
         if (p_l->locus_typ == TLOC_LOC && INHER_SELCTD)
            mak_allefrq (NULL, NULL, 0);
         else
            mak_allefrq (p_m, p_l, p_l->no);

      if (UNLTLOC_SELCTD)               /* or unlinked tloc                   */
      {
         entry = FALSE;
         while (tl_no = nxt_sim_unln_tl_no (&entry))
            mak_allefrq (NULL, NULL, tl_no);
      }
      else if (!NTLocs)                 /* or no tloc                         */
         mak_allefrq (NULL, NULL, 0);
   }

   if (MARK_NAME)
   {
      if (!test_marknam (p_s))          /* checks marker names                */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         if (mak_marknam (p_s))         /* stores and prints marker names     */
            prnt_marknam (p_s);         /* if provided                        */
      }
      else
         return proc_end ();
   }

   if (MARK_DATA)
   {
      if (!test_markdat (p_s))          /* checks marker data                 */
         if (proc_err ())
            return FALSE;
      if (!n_err)
      {
         mak_markdat (p_s);             /* stores and prints marker data      */
         prnt_markdat (p_s);
      }
      else
         return proc_end ();
   }

   if (!test_lods_pars ())              /* checks MC computational parametrs */
      if (proc_err ())
         return FALSE;

   if (TRAIT_DATA)                      /* checks input record parmeters     */
      if (!test_ped_trtdat ())
         if (proc_err ())
            return FALSE;
   if (n_err)
      return proc_end ();

   mak_ped_dat ();

   if (TRAIT_DATA)                      /* sets up and reports trait ped data */
   {
      prnt_trt_typ ();

      prnt_ped_dat ();
   }

   if (TRAIT_PENE)                      /* checks incomplete penetrances      */
   {                                    /* for discete trait data             */
      if (!test_penetrnc ())
         if (proc_err ())
            return FALSE;

      prnt_penetrnc ();                 /* and prints penetrances             */
   }

   if (n_err)
      return proc_end ();

   if (NTraits && (TraitTyp == QUANTITATIVE)) /* stores geno means & variance */
   {
      alloc_traits ();
      entry = FALSE;
      tl_no = nxt_tr_tl_no (&entry, t_no);
      loc_no = get_tl_sim_loc_no (tl_no);
      mak_trait (t_no, tl_no, loc_no, 0);
      prnt_qtrait_pars (p_s);           /* prints means and variances         */
   }

   prnt_lods_pars ();                   /* prints MC parameters               */

#endif

   return TRUE;
}
